#!/usr/bin/perl -wT
#
# Copyright (c) 2017 University of Utah and the Flux Group.
# 
# {{{EMULAB-LICENSE
# 
# This file is part of the Emulab network testbed software.
# 
# This file is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
# 
# This file is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this file.  If not, see <http://www.gnu.org/licenses/>.
# 
# }}}
#
# Implements the libvnode API for Docker support in Emulab.
#
# Note that there is no distinguished first or last call of this library
# in the current implementation.  Every vnode creation (through mkvnode.pl)
# will invoke all the root* and vnode* functions.  It is up to us to make
# sure that "one time" operations really are executed only once.
#
package dockerclient;
use Exporter;
@ISA    = "Exporter";
@EXPORT = qw( new
    
);
use vars qw();

use strict;
use warnings;
use Data::Dumper;
use LWP::UserAgent;
use LWP::Protocol::http::SocketUnixAlt;
use JSON::PP;
use URI::Escape;
use MIME::Base64 qw(encode_base64url encode_base64);

#
# Turn off line buffering on output
#
$| = 1;

my $DOCKSOCK = "/var/run/docker.sock";

#
# Seems like we cannot just instantiate a unix sock protocol and pass it
# to a UserAgent; we have to set a "protocol implementor".  So we use a
# fake protocol to ensure we don't collide with a legit protocol.
#
LWP::Protocol::implementor(dlhttp => 'LWP::Protocol::http::SocketUnixAlt');

sub New($;$) {
    my ($class,$sockpath) = @_;

    $sockpath = $DOCKSOCK
	if (!defined($sockpath));
    $self->{"sockpath"} = $sockpath;
    $self->{"ua"} = LWP::UserAgent->new();

    bless($self,$class);

    return $self;
}

sub sockpath($) {
    my ($self,) = @_;
    return $self->{"sockpath"};
}

sub request($$) {
    my ($self,$req) = @_;

    return $self->{"ua"}->request($req);
}

sub post($$$;$) {
    my ($self,$uri,$headers,$content) = @_;

    my $req = HTTP::Request(
	"POST","dlhttp:".$self->sockpath()."$uri",$headers,$content);
    my $res = self->request($req);
    if ($res->is_success) {
	return (0,$res->content,$res);
    }
    else {
	return ($res->code,$res->content,$res);
    }
}

sub get($$$;$) {
    my ($self,$uri,$headers,$content) = @_;

    my $req = HTTP::Request(
	"GET","dlhttp:".$self->sockpath()."$uri",$headers,$content);

    my $res = self->request($req);
    if ($res->is_success) {
	return (0,$res->content,$res);
    }
    else {
	return ($res->code,$res->content,$res);
    }
}

sub head($$$;$) {
    my ($self,$uri,$headers,$content) = @_;

    my $req = HTTP::Request(
	"HEAD","dlhttp:".$self->sockpath()."$uri",$headers,$content);
    my $res = self->request($req);
    if ($res->is_success) {
	return (0,$res->content,$res);
    }
    else {
	return ($res->code,$res->content,$res);
    }
}

sub delete($$$;$) {
    my ($self,$uri,$headers,$content) = @_;

    my $req = HTTP::Request(
	"DELETE","dlhttp:".$self->sockpath()."$uri",$headers,$content);
    my $res = self->request($req);
    if ($res->is_success) {
	return (0,$res->content,$res);
    }
    else {
	return ($res->code,$res->content,$res);
    }
}

sub Commit($$$$) {
    my ($self,$container,$repo,$tag) = @_;

    my $erepo = uri_escape($repo);
    my $etag = uri_escape($tag);
    
    return $self->post("/commit?container=$container&repo=$erepo&tag=$etag");
}

sub Pull($$;$$) {
    my ($self,$image,$user,$pass) = @_;

    my $headers = {};
    if (defined($user) && $user ne "") {
	# Default to the default registry if the image doesn't have a
	# host:port part.
	my $registry = "registry-1.docker.io";
	if ($image =~ /^([a-zA-Z0-9-\.]+)(:\d+)?\/.*$/) {
	    $registry = "$1$2";
	}
	my $auth = encode_base64url(
	    '{"username":"'.$user.'","password":"'.$pass.'"}'."");
	chomp($auth);
	$auth =~ tr/\r\n//;
	$headers->{'X-Registry-Auth'} = $auth;
    }
    my $uimage = uri_escape($image);

    return $self->post("/images/create?fromImage=$uimage",$headers);
}

sub ConnectContainerToNetwork($$$$$) {
    my ($self,$container,$netid,$ip,$maskbits,$macaddr) = @_;

    my $headers = { "Content-Type" => "application/json" };
    my $data = \
	'{"Container": "'.$container'","'.
	' "EndpointConfig":'.
	'   {"IPAMConfig":{"IPv4Address":"'.$ip.'","'.
	'    "IPPrefixLen":'.$maskbits.',"Gateway":""},'.
	'    "MacAddress":"'.$macaddr'"}}}';

    return $self->post("/networks/$netid/connect",$headers,$data);
}

sub NetworkExists($$) {
    my ($self,$netid) = @_;

    my ($code,) = $self->get("/networks/$netid");

    return $code;
}

sub CreateBridgeNetwork($$$$$;$) {
    my ($self,$name,$cidr,$gateway,$brname) = @_;

    my $data = {
	"Name" => $name,"Driver" => "bridge",
	"IPAM" => { "Config" => { "Subnet" => $cidr,"Gateway" => $gateway } }
    };
    if (defined($brname)) {
	$data->{"Options"} = { "com.docker.network.bridge.name" => $brname };
    }
    return $self->post("/networks/create",{},encode_json($data));
}

sub CreateMacvlanNetwork($$$$$;$) {
    my ($self,$name,$cidr,$gateway,$basedev) = @_;

    my $data = {
	"Name" => $name,"Driver" => "macvlan",
	"IPAM" => { "Config" => { "Subnet" => $cidr,"Gateway" => $gateway } }
    };
    if (defined($basedev)) {
	$data->{"Options"} = { "parent" => $basedev };
    }
    return $self->post("/networks/create",{},encode_json($data));
}

sub DeleteNetwork($$) {
    my ($self,$netid) = @_;

    return $self->delete("/networks/$netid");
}

sub GetContainer($$) {
    my ($self,$id) = @_;

    my ($code,$resp,$res) = self->get("/containers/$id/json");
    if ($code) {
	return ($code,$resp,$res);
    }
    else {
	return ($code,decode_json($resp),$res);
    }
}

sub GetContainerStatus($$) {
    my ($self,$id) = @_;

    my ($code,$resp,$res) = $self->GetContainer($id);
    if ($code) {
	return ($code,$resp,$res);
    }
    else {
	$resp = decode_json($resp);
	return (0,$resp{"State"}{"Status"},$res);
    }
}

